import os
import urllib
import zipfile
from os.path import expanduser
import logging

model_with_visual_url = "https://figshare.com/ndownloader/files/34494836"
model_without_visual_url = "https://figshare.com/ndownloader/files/34494839"


def get_model_path(visual=False):
    if visual:
        model_name = "h2v_1000_100"
        model_url = model_with_visual_url
    else:
        model_name = "h2v_1000_100_text_only"
        model_url = model_without_visual_url
    model_home = os.path.join(expanduser("~"), '.homepage2vec')
    os.makedirs(model_home, exist_ok=True)
    model_folder = os.path.join(model_home, model_name)
    if not os.path.exists(model_folder):
        logging.debug('Downloading model {} in {}'.format(model_name, model_folder))
        filename, headers = urllib.request.urlretrieve(model_url)
        with zipfile.ZipFile(filename, "r") as zip_ref:
            zip_ref.extractall(model_home)
        logging.debug('Downloading model - Done.')
    else:
        logging.debug('Model {} available in {}'.format(model_name, model_folder))
    return model_home, model_folder



import logging

import torch
from torch import nn
from torch.nn import functional as F
import numpy as np
from homepage2vec.model_loader import get_model_path
from homepage2vec.textual_extractor import TextualExtractor
from homepage2vec.visual_extractor import VisualExtractor
from homepage2vec.data_collection import access_website, take_screenshot
import uuid
import tempfile
import os
from os.path import expanduser
import glob
import json
import urllib.request
import zipfile


class WebsiteClassifier:
    """
    Pretrained Homepage2vec model
    """

    def __init__(self, visual=False, device=None, cpu_threads_count=1, dataloader_workers=1):
        self.input_dim = 5177 if visual else 4665
        self.output_dim = 14
        self.classes = ['Arts', 'Business', 'Computers', 'Games', 'Health', 'Home', 'Kids_and_Teens',
                        'News', 'Recreation', 'Reference', 'Science', 'Shopping', 'Society', 'Sports']

        self.with_visual = visual

        self.model_home_path, self.model_path = get_model_path(visual)

        self.temporary_dir = tempfile.gettempdir() + "/homepage2vec/"
        os.makedirs(self.temporary_dir + "/screenshots", exist_ok=True)
        # clean screen shorts
        files = glob.glob(self.temporary_dir + "/screenshots/*")
        for f in files:
            os.remove(f)

        self.device = device
        self.dataloader_workers = dataloader_workers
        os.environ["TOKENIZERS_PARALLELISM"] = "false"
        if not device:
            if torch.cuda.is_available():
                self.device = 'cuda:0'
            else:
                self.device = 'cpu'
                torch.set_num_threads(cpu_threads_count)

        # load pretrained model
        model_tensor = torch.load(self.model_path + "/model.pt", map_location=torch.device(self.device))
        self.model = SimpleClassifier(self.input_dim, self.output_dim)
        self.model.load_state_dict(model_tensor)

        # features used in training
        self.features_order = []
        self.features_dim = {}
        logging.debug("Loading features from {}".format(self.model_path + '/features.txt'))
        with open(self.model_path + '/features.txt', 'r') as file:
            for f in file:
                name = f.split(' ')[0]
                dim = int(f.split(' ')[1][:-1])
                self.features_order.append(name)
                self.features_dim[name] = dim

    def get_scores(self, x):
        with torch.no_grad():
            self.model.eval()
            return self.model.forward(x)

    def fetch_website(self, url):
        logging.debug("Fetching website: {}".format(url))
        response = access_website(url)
        w = Webpage(url)
        if response is not None:
            html, get_code, content_type = response
            w.http_code = get_code
            if self.is_valid(get_code, content_type):
                w.is_valid = True
                w.html = html
        if self.with_visual:
            logging.debug("Generating screenshot: {}".format(url))
            out_path = self.temporary_dir + "/screenshots/" + str(w.uid)
            w.screenshot_path = take_screenshot(w.url, out_path)
            logging.debug("Screenshot for {} ready in {}".format(url, w.screenshot_path))
        return w

    def get_features(self, url, html, screenshot_path):
        te = TextualExtractor(self.device)
        features = te.get_features(url, html)
        if self.with_visual:
            ve = VisualExtractor(self.device)
            visual_features = ve.get_features(screenshot_path)
            features['f_visual'] = visual_features
        return features

    def predict(self, website):
        website.features = self.get_features(website.url, website.html, website.screenshot_path)
        all_features = self.concatenate_features(website)
        input_features = torch.FloatTensor(all_features)
        scores, embeddings = self.get_scores(input_features)
        return dict(zip(self.classes, torch.sigmoid(scores).tolist())), embeddings.tolist()


    def concatenate_features(self, w):
        """
        Concatenate the features attributes of webpage instance, with respect to the features order in h2v
        """

        v = np.zeros(self.input_dim)

        ix = 0

        for f_name in self.features_order:
            f_dim = self.features_dim[f_name]
            f_value = w.features[f_name]
            if f_value is None:
                f_value = f_dim * [0]  # if no feature, replace with zeros
            v[ix:ix + f_dim] = f_value
            ix += f_dim

        return v

    def is_valid(self, get_code, content_type):
        valid_get_code = get_code == 200
        valid_content_type = content_type.startswith('text/html')
        return valid_get_code and valid_content_type


class SimpleClassifier(nn.Module):
    """
    Model architecture of Homepage2vec
    """

    def __init__(self, input_dim, output_dim, dropout=0.5):
        super(SimpleClassifier, self).__init__()

        self.layer1 = torch.nn.Linear(input_dim, 1000)
        self.layer2 = torch.nn.Linear(1000, 100)
        self.fc = torch.nn.Linear(100, output_dim)

        self.drop = torch.nn.Dropout(dropout)  # dropout of 0.5 before each layer

    def forward(self, x):
        x = self.layer1(x)
        x = F.relu(self.drop(x))

        emb = self.layer2(x)
        x = F.relu(self.drop(emb))

        x = self.fc(x)

        return x, emb


class Webpage:
    """
    Shell for a webpage query
    """

    def __init__(self, url):
        self.url = url
        self.uid = uuid.uuid4().hex
        self.is_valid = False
        self.http_code = False
        self.html = None
        self.screenshot_path = None
        self.features = None
        self.embedding = None
        self.scores = None

    def __repr__(self):
        return json.dumps(self.__dict__)



import torch 
import torchvision.models as models
from torchvision import datasets, transforms
from torch import nn
from PIL import Image

###################### CVM ######################

out_dim = 14 # number of classes
features_dim = 512 # number of features before the classifier

class ResNetPretrained(nn.Module):
    def __init__(self):
        super(ResNetPretrained, self).__init__()
        
        resnet = models.resnet18(pretrained=True)
        
        self.features = torch.nn.Sequential(*list(resnet.children())[:-1])       
        self.fc1 = torch.nn.Linear(features_dim, out_dim)


    def forward(self, x):
        x = self.features(x).reshape(-1, features_dim)

        x = self.fc1(x)
        return x
    
    
###################### FEATURES EXTRACTION ######################


class VisualExtractor:
    """
    Extract visual features from the screenshot of a webpage
    """
    
    def __init__(self, device='cpu'):
        self.model = ResNetPretrained().to(device)
        self.device = device
        self.batch_size = 128
       
    
    def get_features(self, screenshot_path):
        
        # dimension of the images
        valid_xdim = 640 
        valid_ydim = 360 

        # factor for 5-crop transform
        crop_factor = 0.6 

        crop_dim = [int(crop_factor * valid_ydim), int(crop_factor * valid_xdim)]

        # 5-crop transform
        five_crop = transforms.FiveCrop(size=crop_dim) 

        # image to tensor transform
        tensorize = transforms.ToTensor() 

        # normalization transform
        normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]) 

        # to apply to each crop
        stack_norm_tensorize = transforms.Lambda(
            lambda crops: torch.stack([normalize(tensorize(crop)) for crop in crops]))

        # final transform: crop > tensorize > normalize
        data_transforms = transforms.Compose([five_crop, stack_norm_tensorize])

        img = Image.open(screenshot_path)
        img_tensor = data_transforms(img)

        self.model.eval()
        with torch.no_grad():
            inputs = img_tensor.to(self.device)
            ncrops, c, h, w = inputs.size()
            outputs = self.model.features(inputs.view(-1, c, h, w)) # output for each crop
            outputs = outputs.view(1, ncrops, -1).mean(1) # mean over the crops
        return outputs.detach()[0]




import requests

import time

from selenium import webdriver
from selenium.common.exceptions import WebDriverException

from PIL import Image
import os

import signal
from contextlib import contextmanager


class TimeoutException(Exception): pass


@contextmanager
def time_limit(seconds):
    """ Set a time limit on the execution of a block"""

    def signal_handler(signum, frame):
        raise TimeoutException("Timed out!")

    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)


def access_website(url, timeout=10):
    """
    Return the response corresponding to a url, or None if there was a request error
    """

    try:
        # avoid the script to be blocked
        with time_limit(10 * timeout):

            # change user-agent so that we don't look like a bot
            headers = requests.utils.default_headers()
            headers.update({
                'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.16; rv:84.0) Gecko/20100101 Firefox/84.0',
            })

            # r_head = requests.head("http://" + url, timeout=timeout, headers=headers)
            if not url.startswith("http://") and not url.startswith("https:"):
                url = "http://"+url
            r_get = requests.get(url, timeout=timeout, headers=headers)

            # head_code = r_head.status_code
            get_code = r_get.status_code
            if r_get.encoding.lower() != 'utf-8':
                r_get.encoding = r_get.apparent_encoding
            text = r_get.text
            content_type = r_get.headers.get('content-type', '?').strip()
            return text, get_code, content_type

    except Exception as e:
        return None


def take_screenshot(url, out_path, in_width=1920, in_height=1080, down_factor=3, quality=85, timeout=30):
    """
    Take a screenshot of a website and save it under the outpath
    """

    out_width = int(in_width / down_factor)
    out_height = int(in_height / down_factor)

    try:

        # driver
        options = webdriver.ChromeOptions()
        options.headless = True

        driver = webdriver.Chrome('chromedriver', options=options)

        driver.set_page_load_timeout(timeout)
        driver.set_window_size(in_width, in_height)

        # access the url
        if not url.startswith("http://") and not url.startswith("https:"):
            url = "http://" + url
        driver.get(url)

        # set the opacity to 0 for elements that might be popup, etc...
        try:
            targets = ["popup", "modal", "cookie"]  # the substrings in the div we want to hide
            target_types = ["class", "id"]  # where the substrings are
            js_script = ""
            for tar in targets:
                for ty in target_types:
                    js_script += "document.styleSheets[0].insertRule('div[" + ty + "*=" + tar + \
                                 "] {opacity: 0 !important}', 0); \n"

            driver.execute_script(js_script)

        # if can't access the css sheet
        except WebDriverException as e:
            pass

        # so that the website's elements are loaded
        time.sleep(2)

        # takes a screenshot (only in png)
        driver.save_screenshot(out_path + '.png')

        # convert the png into a jpeg of lesser dimensions and quality
        img = Image.open(out_path + '.png')
        img = img.convert('RGB')
        img = img.resize((out_width, out_height), Image.ANTIALIAS)
        img.save(out_path + '.jpeg', optimize=True, quality=quality)
        os.remove(out_path + '.png')
        return out_path + '.jpeg'

    except Exception as e:
        print(e)
        return

    finally:
        if 'driver' in locals():
            driver.quit()



import re
from bs4 import BeautifulSoup
from sentence_transformers import SentenceTransformer
# import pandas as pd
from collections import Counter

class TextualExtractor:
    """
    Extract textual features from the html content of a webpage
    """

    xlmr = None

    def __init__(self, device='cpu'):
        if not TextualExtractor.xlmr:
            TextualExtractor.xlmr = SentenceTransformer('paraphrase-xlm-r-multilingual-v1', device=device)
        # self.xlmr = SentenceTransformer('xlm-r-distilroberta-base-paraphrase-v1', device=device)
        
        # TLD used for one-hot encoding
        self.rep_tld = ['com', 'org', 'net', 'info', 'xyz', 'club', 'biz', 'top', 'edu', 'online', 
                        'pro', 'site', 'vip', 'icu', 'buzz', 'app', 'asia', 'su', 'gov', 'space']
        
        # Metatags used for one-hot encoding
        self.rep_metatags = ['viewport', 'description', 'generator', 'keywords', 'robots', 'twitter:card', 
                             'msapplication-tileimage', 'google-site-verification', 'author', 'twitter:title', 
                             'twitter:description', 'theme-color', 'twitter:image', 'twitter:site', 
                             'format-detection', 'msapplication-tilecolor', 'copyright', 'twitter:data1', 
                             'twitter:label1', 'revisit-after', 'apple-mobile-web-app-capable', 'handheldfriendly', 
                             'language', 'msvalidate.01', 'twitter:url', 'title', 'mobileoptimized', 
                             'twitter:creator', 'skype_toolbar', 'rating']
        
        # number of sentences and links over which we compute the features
        self.k_sentences = 100
        self.k_links = 50

    def get_features(self, url, html):

        features = {}

        # url
        url_feature = embed_url(url, TextualExtractor.xlmr)
        features['f_url'] = url_feature

        # tld 
        tld_feature = embed_tld(url, self.rep_tld)
        features['f_tld'] = tld_feature

        # print(html)
        soup = BeautifulSoup(html, 'lxml')

        # metatags
        metatags_feature = embed_metatags(soup, self.rep_metatags)
        features['f_metatags'] = metatags_feature

        # title
        title_feature = embed_title(soup, TextualExtractor.xlmr)
        features['f_title'] = title_feature

        # description
        description_feature = embed_description(soup, TextualExtractor.xlmr)
        features['f_description'] = description_feature

        # keywords
        keywords_feature = embed_keywords(soup, TextualExtractor.xlmr)
        features['f_keywords'] = keywords_feature

        # links
        links_feature = embed_links(soup, TextualExtractor.xlmr, self.k_links)
        features['f_links_' + str(self.k_links)] = links_feature

        # text
        text_feature = embed_text(soup, TextualExtractor.xlmr, self.k_sentences)
        features['f_text_' + str(self.k_sentences)] = text_feature

        return features

    

def embed_text(soup, transformer, k_sentences):

    sentences = split_in_sentences(soup)[:k_sentences]

    if len(sentences) == 0:
        return None

    # this is needed to avoid some warnings, truncate the sentences
    sentences_trunc = [trunc(s, transformer.tokenizer, transformer.max_seq_length) for s in sentences]

    sentences_emb = transformer.encode(sentences_trunc)

    if sentences_emb.size == 0:
        return None

    text_emb = sentences_emb.mean(axis=0).tolist() # mean of the sentences 

    return text_emb


def embed_description(soup, transformer):

    desc = soup.find('meta', attrs = {'name': ['description', 'Description']})

    if not desc:
        return None

    content = desc.get('content', '')

    if len(content.strip()) == 0:
        return None

    content = clean_field(content)

    # this is needed to avoid some warnings
    desc_trunc = trunc(content, transformer.tokenizer, transformer.max_seq_length)
    desc_emb = transformer.encode(desc_trunc)

    if desc_emb.size == 0:
        return None

    return desc_emb.tolist()


def embed_keywords(soup, transformer):

    kw = soup.find('meta', attrs = {'name': 'keywords'})

    if not kw: 
        return None

    content = kw.get('content', '')

    if len(content.strip()) == 0:
        return None

    # this is needed to avoid some warnings
    kw_trunc = trunc(content, transformer.tokenizer, transformer.max_seq_length)
    kw_emb = transformer.encode(kw_trunc)

    if kw_emb.size == 0:
        return None

    return kw_emb.tolist()


def embed_title(soup, transformer):

    title = soup.find('title')

    if title is None:
        return None

    title = str(title.string)
    title = clean_field(title)

    if len(title) == 0:
        return None

    # this is needed to avoid some warnings
    title_trunc = trunc(title, transformer.tokenizer, transformer.max_seq_length)
    title_emb = transformer.encode(title_trunc)

    if title_emb.size == 0:
        return None

    return title_emb.tolist()


def embed_links(soup, transformer, k_links):

    a_tags = soup.find_all('a', href=True)

    links = [a.get('href', '') for a in a_tags]
    links = [clean_link(link) for link in links]
    links = [link for link in links if len(link) != 0]

    words = [w.lower() for w in ' '.join(links).split(' ') if len(w) != 0]

    if len(words) == 0:
        return None

    most_frequent_words = [w[0] for w in Counter(words).most_common(k_links)]

    # most_frequent_words = pd.Series(words).value_counts()[:k_links].index.values

    # this is needed to avoid some warnings
    words_trunc = [trunc(w, transformer.tokenizer, transformer.max_seq_length) for w in most_frequent_words]
    words_emb = transformer.encode(words_trunc)

    if words_emb.size == 0:
        return None

    links_emb = words_emb.mean(axis=0).tolist()

    return links_emb


def embed_url(url, transformer):

    cleaned_url = clean_url(url)

    # this is needed to avoid some warnings
    url_trunc = [trunc(w, transformer.tokenizer, transformer.max_seq_length) for w in cleaned_url]
    url_emb = transformer.encode(cleaned_url)

    if url_emb.size == 0:
        return None

    return url_emb.mean(axis=0).tolist()



def embed_tld(url, rep_tld):

    tld = url.split('.')[-1]
    rep_onehot = [int(tld.startswith(d)) for d in rep_tld]
    continent_onehot = 7*[0] #TODO

    return rep_onehot + continent_onehot



def embed_metatags(soup, rep_metatags):

    metatags = soup.findAll('meta')
    attr = [m.get('name', None) for m in metatags]
    attr = [a.lower() for a in attr if a != None]

    attr_emb = [int(a in attr) for a in rep_metatags]

    return attr_emb



def split_in_sentences(soup):
    """ From the raw html content of a website, extract the text visible to the user and splits it in sentences """

    sep = soup.get_text('[SEP]').split('[SEP]') # separate text elements with special separators [SEP]
    strip = [s.strip() for s in sep if s != '\n']
    clean = [s for s in strip if len(s) != 0]

    return clean


def clean_url(url):
    url = re.sub(r"www.|http://|https://|-|_", '', url)
    return url.split('.')[:-1]


def clean_field(field):
    field = re.sub(r"\*|\n|\r|\t|\||:|-|–", '', field)
    return field.strip()


def clean_link(link):
    link = re.sub(r"www.|http://|https://|[0-9]+", '', link)
    link = re.sub(r"-|_|=|\?|:", ' ', link)
    link = link.split('/')[1:]
    return ' '.join(link).strip()


def trunc(seq, tok, max_length):
    """ Truncate the output of a tokenizer to a given length, doesn't affect the performances """
    e = tok.encode(seq, truncation=True)
    d = tok.decode(e[1:-1][:max_length-2])
    return d

